/*
 *  Tcpjunk - TCP protocols testing tool
 *  Contact: oferkv@gmail.com
 *
 *  Tcpjunk is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2, or (at your option) any
 *  later version.
 *
 *  Tcpjunk is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/wait.h>
#include <netdb.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <netinet/tcp.h>
#include "tcpjunk.h"

void show_help()
  {
  printf("Tcpjunk v%s - TCP protocols testing tool\n", VER);
  puts(
      "\nGeneral options:\n\
    -c <filename>  session setup file\n\
    -f             use random session files from current directory\n\
    -d             force display of session data when using instances\n\
                   or in server mode\n\
    -v             display warnings and other information\n\
    -h             show this help and exit\n\
    -x             tcpjunk in GUI mode\n\n\
Client options:\n\
    -i <number>    number of instances to run in parallel\n\
    -l <number>    number of loops for each instance (default: forever)\n\
    -s <hostname>  hostname/IP of target\n\
    -P <IP>        use specific source IP address\n\
    -R             use random source IP addresses\n\
    -I <interface> source IP alias interface\n\
    -o <offset>    source IP interface alias offset\n\
    -O <port>      bind to specific source port\n\n\
Server options:\n\
    -r             server mode\n\
    -E <filename>  server Certificate file (default: server.crt)\n\
    -K <filename>  server key file (default: server.key)\n\n\
Connection options:\n\
    -p <port>      TCP port to connect/listen to\n\
    -W             wait for data at first\n\
    -C             do not wait for data at last\n\
    -w <mseconds>  delay between connections (default: none)\n\
    -k <times>     repeat on the same connection (default: 1)\n\
    -t <sec>       initial timeout (default: 2 seconds)\n\
    -T <ms>        reply timeout (default: 100 milliseconds)\n\
    -m <bytes>     send/receive buffer size\n\
    -D <ms>        delay before sending each buffer\n\
    -u             optimize TCP tw_recycle, tw_reuse settings and\n\
                   file descriptor limits\n\
    -S             use SSL v3.0\n\n\
Session setup tags:\n\
   <sect>[data]</sect>            delimiters for multiple sections\n\
   <rep [min] [max]>[data]</rep>  data repeater\n\
   <sys>[command]<sys>            insert output of a command\n\
   <fuzz [type] [min] [max]>      insert a fuzzed buffer\n\
   <char [hex code]>              insert a character\n\
   <inst>                         insert client instance number\n\n\
Fuzzing types:\n\
   digit    [0-9]\n\
   alnum    [0-9][a-z][A-Z]\n\
   alpha    [a-z][A-Z]\n\
   cntrl    control characters\n\
   print    any printable character\n\
   special  above ASCII 127\n\
   punct    punctuation characters\n\
   space    space characters\n\
   any      any character\n\
   random   random fuzzing type for each instance\n\n\
Bug reports and suggestions: <oferkv@gmail.com>");

  exit(EXIT_FAILURE);
  }

int read_reply(int sock, SSL *ssl, struct tj_data *instance_data)
  {
  int sel_ret, read_ret;
  int read_success = 0;
  char read_buf[SOCK_BUF_LEN];

  /* initialize descriptors set */
  fd_set set;
  FD_ZERO(&set);
  FD_SET(sock, &set);

  /* initialize timeout */
  struct timeval timeout;
  timeout.tv_sec = instance_data->itimeout;
  timeout.tv_usec = 0;

  while ((sel_ret = select(sock + 1, &set, NULL, NULL, &timeout))> 0
  || (instance_data->use_ssl && SSL_pending(ssl)))
    {
    /* clean buffer */
    memset(read_buf, 0, sizeof(read_buf));

    /* read clear text */
    if (instance_data->use_ssl == 0)
      {
      read_ret = recv(sock, read_buf, SOCK_BUF_LEN, MSG_DONTWAIT);
      if (read_ret < 0)
        {
        tj_error(instance_data, "Error reading data from socket: ", strerror(
            errno), 1);
        break;
        }
      }
    /* read using SSL */
    else
      {
      read_ret = SSL_read(ssl, read_buf, SOCK_BUF_LEN);
      if (read_ret < 0)
        {
        tj_error(instance_data, "SSL read ", ERR_error_string(ERR_get_error(),
            NULL), 1);
        break;
        }
      }

    /* if no data break */
    if (read_ret == 0)
    break;

    /* print reply */
    if (instance_data->debug> 0 && instance_data->gui == 0)
    fwrite(read_buf, sizeof(char), read_ret, stdout);
    else if (instance_data->gui == 1)
    gui_show_traffic(read_buf, read_ret, 1);

    if (instance_data->verbose)
      {
      char read_len_str[64];
      sprintf(read_len_str, "Received %d bytes", read_ret);
      tj_error(instance_data, read_len_str, "", 0);
      }

    if (!read_success)
    read_success = 1;

    /* reset timeouts */
    timeout.tv_sec = 0;
    timeout.tv_usec = instance_data->stimeout;

    }

  if (read_success)
  return 0;
  else
  return 1;

  }

void send_data(int sock, SSL *ssl, struct tj_data *instance_data,
    char *send_buf, int send_buf_len)
  {

  int rc;
  int err;

  if (instance_data->use_ssl == 0)
    {

    /* send junk clear text */
    if (instance_data->segment_size == 0)
      {
      rc = send(sock, send_buf, send_buf_len, 0);
      if (rc < 0)
        tj_error(instance_data, "Error sending data: ", strerror(errno), 1);
      }
    else
      {
      /* segment size */
      int loc, seg_len = instance_data->segment_size;
      for (loc = 0; loc < send_buf_len; loc += seg_len)
        {
        if ((send_buf_len - loc) < seg_len)
          seg_len = send_buf_len - loc;
        rc = send(sock, &send_buf[loc], seg_len, 0);
        if (instance_data->segment_delay > 0)
          usleep(instance_data->segment_delay);
        if (rc < 0)
          {
          tj_error(instance_data, "Error sending data: ", strerror(errno), 1);
          break;
          }
        }
      }
    }
  else
    {
    /* send junk using SSL */

    if (instance_data->segment_size == 0)
      {
      err = SSL_write(ssl, send_buf, send_buf_len);
      if ((err) == -1)
        tj_error(instance_data, "SSL write ", ERR_error_string(ERR_get_error(),
            NULL), 1);
      }
    else
      {
      /* buffer size limit */
      int loc, seg_len = instance_data->segment_size;
      for (loc = 0; loc < send_buf_len; loc += seg_len)
        {
        if ((send_buf_len - loc) < seg_len)
          seg_len = send_buf_len - loc;
        err = SSL_write(ssl, &send_buf[loc], seg_len);
        if (instance_data->segment_delay > 0)
          usleep(instance_data->segment_delay);
        if ((err) == -1)
          tj_error(instance_data, "SSL write ", ERR_error_string(
              ERR_get_error(), NULL), 1);
        }
      }
    }

  /* print sent buffer if we are debugging */
  if (instance_data->debug > 0 && instance_data->gui == 0)
    fwrite(send_buf, sizeof(char), send_buf_len, stdout);
  else if (instance_data->gui == 1)
    {
    gui_show_traffic(send_buf, send_buf_len, 0);
    }
  if (instance_data->verbose)
    {
    char read_len_str[64];
    sprintf(read_len_str, "Sent %d bytes", send_buf_len);
    tj_error(instance_data, read_len_str, "", 0);
    }

  }

int tj_connect(int sock, struct sockaddr_in sa, struct tj_data *instance_data)
  {
  int ret = 0, sock_opts;

  /* set non-blocking on socket */
  if ((sock_opts = fcntl(sock, F_GETFL, 0)) >= 0)
    ret = fcntl(sock, F_SETFL, sock_opts | O_NONBLOCK);
  if (ret < 0)
    return -1;

  /* buffer size limit */
  if (instance_data->segment_size)
    {
    int flag = 1;
    ret = setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, (char *) &flag,
        sizeof(int));
    if (ret < 0)
      tj_error(instance_data, "Disabling TCP_NODELAY failed: ",
          strerror(errno), 1);
    }

  /* connect */
  ret = connect(sock, (struct sockaddr *) &sa, sizeof(sa));
  if ((ret < 0) && (errno!= EINPROGRESS))
    return -1;

  /* initialize descriptors set */
  fd_set set;
  FD_ZERO(&set);
  FD_SET(sock, &set);

  /* initialize timout */
  struct timeval timeout;
  timeout.tv_sec = instance_data->itimeout;
  timeout.tv_usec = 0;

  ret = select(sock + 1, NULL, &set, NULL, &timeout);
  if (ret > 0)
    {
    int so_ret, get_ret;
    unsigned int get_len = sizeof(get_ret);

    /* handle errors */
    so_ret = getsockopt(sock, SOL_SOCKET, SO_ERROR, &get_ret, &get_len);
    if (so_ret < 0)
      {
      tj_error(instance_data, "System request failed: ", strerror(errno), 1);
      return -1;
      }
    if (get_ret > 0)
      {
      errno = get_ret;
      return -1;
      }

    }
  else if (ret <= 0)
    {
    errno = ETIMEDOUT;
    return -1;
    }

  /* all was good, restore socket options */
  ret = fcntl(sock, F_SETFL, sock_opts);

  if (instance_data->verbose == 1)
    tj_error(instance_data, "Connected", "", 0);
  return 0;
  }

void tj_disconnect(int sock, struct tj_data *instance_data, SSL * ssl,
    SSL_CTX * ctx)
  {

  int err = 0;
  if (instance_data->use_ssl == 1)
    {
    if (ssl != NULL)
      err = SSL_clear(ssl);
    if ((err) == -1)
      tj_error(instance_data, "SSL Clear ", ERR_error_string(ERR_get_error(),
          NULL), 1);
    }

  shutdown(sock, SHUT_RDWR);
  err = close(sock);
  if (err < 0)
    tj_error(instance_data, "Closing socket failed: ", strerror(errno), 1);
  else if (instance_data->verbose == 1)
    tj_error(instance_data, "Connection closed", "", 0);

  }

long tj_filesize(FILE *stream)
  {
  long curpos, len;
  curpos = ftell(stream);
  fseek(stream, 0L, SEEK_END);
  len = ftell(stream);
  fseek(stream, curpos, SEEK_SET);
  return len;
  }

char fuzz_char(int ftype)
  {

  int rg = 0;
  int tr = 0;

  /* digit 48-57 */
  if (ftype == r_digit)
    return rand() % 10 + 48;

  /* alnum 48-57 65-90 97-122 */
  else if (ftype == r_alnum)
    {
    rg = rand() % 3;
    if (rg == 0)
      return rand() % 10 + 48;
    else if (rg == 1)
      return rand() % 26 + 65;
    else if (rg == 2)
      return rand() % 26 + 97;
    }

  /* alpha 65-90 97-122 */
  else if (ftype == r_alpha)
    {
    rg = rand() % 2;
    if (rg == 0)
      return rand() % 26 + 65;
    else if (rg == 1)
      return rand() % 26 + 97;
    }

  /* cntrl 1-31 127 */
  if (ftype == r_cntrl)
    {
    tr = rand() % 33 + 1;
    if (tr == 32)
      tr = 127;
    return tr;
    }

  /* print 32-126 */
  if (ftype == r_print)
    return rand() % 95 + 32;

  /* special 128-255 */
  if (ftype == r_special)
    return rand() % 128 + 128;

  /* punct 33-47 58-64 91-96 123-126 */
  else if (ftype == r_punct)
    {
    rg = rand() % 4;
    if (rg == 0)
      return rand() % 15 + 33;
    else if (rg == 1)
      return rand() % 7 + 58;
    else if (rg == 2)
      return rand() % 6 + 91;
    else if (rg == 3)
      return rand() % 4 + 123;
    }

  /* space 9-13 32 */
  if (ftype == r_space)
    {
    tr = rand() % 6 + 9;
    if (tr == 14)
      tr = 32;
    return tr;
    }

  /* any 0-255 */
  if (ftype == r_any)
    return rand() % 255;

  /* we never reach here */
  return 0;
  }

/* check fuzz type */
int get_fuzz_type(char *fuzz_type, struct tj_data *instance_data)
  {
  if (strcmp(fuzz_type, "none") == 0)
    return r_none;
  else if (strcmp(fuzz_type, "digit") == 0)
    return r_digit;
  else if (strcmp(fuzz_type, "alnum") == 0)
    return r_alnum;
  else if (strcmp(fuzz_type, "alpha") == 0)
    return r_alpha;
  else if (strcmp(fuzz_type, "cntrl") == 0)
    return r_cntrl;
  else if (strcmp(fuzz_type, "print") == 0)
    return r_print;
  else if (strcmp(fuzz_type, "special") == 0)
    return r_special;
  else if (strcmp(fuzz_type, "punct") == 0)
    return r_punct;
  else if (strcmp(fuzz_type, "space") == 0)
    return r_space;
  else if (strcmp(fuzz_type, "any") == 0)
    return r_any;
  else if (strcmp(fuzz_type, "random") == 0)
    return r_random;
  else
    {
    tj_error(instance_data, "Invalid fuzz type - ", fuzz_type, 1);
    if (instance_data->gui == 0)
      exit(EXIT_FAILURE);
    else
      return 0;
    }

  }

void sigchld_handler(int s)
  {
  while (wait(NULL) > 0)
    ;
  }

void load_session_data(struct tj_data *instance_data)
  {
  /* load session file */
  FILE *fp;
  if ((fp = fopen(instance_data->session_file_name, "r")) == NULL)
    {
    printf("tcpjunk: can not open session file %s\n",
        instance_data->session_file_name);
    exit(EXIT_FAILURE);
    }

  /* set session length */
  instance_data->session_data_len = tj_filesize(fp);

  /* setup file buffer */
  if (instance_data->session_data)
    free(instance_data->session_data);
  instance_data->session_data = tj_malloc(instance_data->session_data,
      instance_data->session_data_len + 1);

  /* read file data */
  fread(instance_data->session_data, instance_data->session_data_len, 1, fp);
  fclose(fp);
  }

int check_address(struct tj_data *instance_data, char *addr)
  {

  if (*addr == '\0')
    goto failure;

  /* check address */
  if (instance_data->verbose)
    tj_error(instance_data, "Resolving address", "", 0);
  struct hostent *hnames;
  hnames = gethostbyname(addr);
  if (!hnames)
    goto failure;
  else
    {
    instance_data->dest_addr = hnames->h_addr_list[0];
    instance_data->dest_addr_len = hnames->h_length;
    if (instance_data->verbose)
      tj_error(instance_data, "Connecting to ", inet_ntoa(
          *(struct in_addr *) hnames->h_addr_list[0]), 0);
    return 1;
    }

  failure:

  tj_error(instance_data, "Invalid address ", addr, 1);
  if (instance_data->gui == 0)
    exit(EXIT_FAILURE);
  else
    return 0;

  }

int check_port(int port, struct tj_data *instance_data)
  {
  /* check port */
  if (port < 1 || port > 65535)
    {
    tj_error(instance_data, "Invalid port", "", 1);
    if (instance_data->gui == 0)
      exit(EXIT_FAILURE);
    else
      return 0;
    }
  return 1;
  }

void tj_error(struct tj_data *instance_data, char *message, char *param,
    int color)
  {
  if (instance_data->debug || instance_data->verbose)
    {
    char temp_err[1024];

    /* print to stderr or in GUI */
    if (instance_data->gui == 0)
      {
      sprintf(temp_err, "%s %s%s\n", "tcpjunk:", message, param);
      fprintf(stderr, "%s", temp_err);
      }
    else
      {
      sprintf(temp_err, "%s%s\n", message, param);
      gui_add_log(temp_err, color);
      }
    }
  }

void find_sections(struct tj_data *instance_data)
  {
  /* find all send buffers in session data */
  char *config_buf = instance_data->session_data;
  instance_data->sect_idx = 0;
  char *section_end;
  char *start_section_tag = "<sect>";
  char *end_section_tag = "</sect>";

  /* If there are sections, enumerate them */
  while ((instance_data->sections[instance_data->sect_idx] = strstr(config_buf,
      start_section_tag)))
    {
    /* start tag found */
    instance_data->sections[instance_data->sect_idx] += strlen(
        start_section_tag);

    /* get section end */
    if (!(section_end = strstr(
        instance_data->sections[instance_data->sect_idx], end_section_tag)))
      {
      /* end tag not found, using all buffer */
      instance_data->sections[instance_data->sect_idx] = config_buf;
      instance_data->sections_len[instance_data->sect_idx]
          = instance_data->session_data_len;
      if (instance_data->sect_idx == 0)
        instance_data->sect_idx++;
      return;
      }
    /* end tag found */
    instance_data->sections_len[instance_data->sect_idx] = section_end
        - instance_data->sections[instance_data->sect_idx];

    instance_data->sect_idx++;
    config_buf = section_end + 1;
    }

  /* just a single section */
  if (instance_data->sect_idx == 0)
    {
    instance_data->sections[instance_data->sect_idx] = config_buf;
    instance_data->sections_len[instance_data->sect_idx]
        = instance_data->session_data_len;
    instance_data->sect_idx++;
    }

  return;

  }

void set_random_session_file(struct tj_data *instance_data)
  {
  strcpy(instance_data->session_file_name, instance_data->rand_file_list[rand()
      % instance_data->rand_files_num]);
  }

void load_rand_file_list(struct tj_data *instance_data)
  {
  int file_index = 0;
  struct dirent* dfile;

  /* open directory */
  DIR *cur_dir = opendir("./");
  if (!cur_dir)
    {
    perror("tcpjunk: error opening directory ");
    exit(EXIT_FAILURE);
    }

  /* get number of files */
  while ((dfile = readdir(cur_dir)) != NULL)
    {
    if (strcmp(dfile->d_name, "..") != 0 && strcmp(dfile->d_name, ".") != 0)
      file_index++;
    }
  instance_data->rand_file_list = (char **) malloc(file_index
      * sizeof(*instance_data->rand_file_list));
  if (instance_data->rand_file_list == NULL)
    {
    printf("tcpjunk: memory allocation failed, terminating...\n");
    exit(EXIT_FAILURE);
    }

  /* store file names */
  int i = 0;
  rewinddir(cur_dir);
  while (((dfile = readdir(cur_dir)) != NULL))
    {
    if (strcmp(dfile->d_name, "..") != 0 && strcmp(dfile->d_name, ".") != 0)
      {
      instance_data->rand_file_list[i] = strdup(dfile->d_name);

      if (instance_data->rand_file_list[i] == NULL)
        {
        printf("tcpjunk: memory allocation failed, terminating...\n");
        exit(EXIT_FAILURE);
        }
      i++;
      }
    }

  /* close */
  if (closedir(cur_dir) == -1)
    {
    perror("tcpjunk: error closing directory ");
    exit(EXIT_FAILURE);
    }
  instance_data->rand_files_num = file_index;

  }

void set_src_ip(struct tj_data *instance_data, char *bind_if_name,
    char *new_src_ip)
  {

  /* setup IP on alias */
  in_addr_t ip = inet_addr(new_src_ip);
  int sock = socket(AF_INET,SOCK_DGRAM, 0);
  struct ifreq ifr;

  strcpy(ifr.ifr_name, bind_if_name);
  ifr.ifr_addr.sa_family = AF_INET;
  ((struct sockaddr_in *) &ifr.ifr_addr)->sin_addr.s_addr = ip;
  if (ioctl(sock, SIOCSIFADDR, &ifr) < 0)
    tj_error(instance_data, "Error setting IP address: ", strerror(errno), 1);

  close(sock);

  }

char *tj_malloc(char *buf, int buf_len)
  {
  buf = malloc(buf_len);
  if (buf == NULL)
    {
    printf("tcpjunk: memory allocation failed, terminating...\n");
    exit(EXIT_FAILURE);
    }
  else
    return buf;
  }

char *tj_realloc(char *buf, int buf_len)
  {
  buf = realloc(buf, buf_len);
  if (buf == NULL)
    {
    printf("tcpjunk: memory allocation failed, terminating...\n");
    exit(EXIT_FAILURE);
    }
  else
    return buf;
  }

void cleanup(int sig)
  {
  printf("tcpjunk: terminating...\n");
  exit(EXIT_SUCCESS);
  }

